#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'

import inspect
import json
import locale
import os

import venusian
from pyramid.exceptions import ConfigurationError
from pyramid.interfaces import IRequest
from z3c.form.action import Action
from z3c.form.browser.button import ButtonWidget
from z3c.form.browser.select import SelectWidget
from z3c.form.browser.submit import SubmitWidget
from z3c.form.browser.text import TextWidget
from z3c.form.browser.textarea import TextAreaWidget
from z3c.form.button import ButtonAction
from z3c.form.converter import BaseDataConverter, DatetimeDataConverter as BaseDatetimeDataConverter, \
    FormatterValidationError
from z3c.form.interfaces import DISPLAY_MODE, IButtonAction, IDataConverter, IFieldWidget, INPUT_MODE, ITextAreaWidget, \
    IWidgetLayoutTemplate, NO_VALUE
from z3c.form.widget import FieldWidget, WidgetLayoutFactory, WidgetTemplateFactory
from zope.component import getGlobalSiteManager
from zope.interface import Interface, directlyProvides, implementer, implementer_only
from zope.pagetemplate.interfaces import IPageTemplate
from zope.schema.fieldproperty import FieldProperty
from zope.schema.interfaces import IBytes, IChoice, IDate, IDatetime, IFloat, IInt, IList, ISet, ITime, ITuple

from pyams_form.interfaces.form import IActionWidget, ICloseWidget, IColorWidget, IDateWidget, IDatetimeWidget, \
    IFloatWidget, IFormLayer, IHTMLWidget, IIntegerWidget, IResetWidget, ISEOTextLineWidget, ISelect2Widget, \
    ITextLineListWidget, ITimeWidget
from pyams_form.schema import IActionButton, ICloseButton, IResetButton
from pyams_skin.interfaces.tinymce import ITinyMCEConfiguration
from pyams_utils.adapter import adapter_config
from pyams_utils.interfaces.data import IObjectData
from pyams_utils.schema import IColorField, IHTMLField, ITextLineListField
from pyams_utils.timezone import localgmtime, tztime

from pyams_form import _


#
# Widget template configuration
#

class widgettemplate_config(object):
    """Class decorator used to declare a widget template"""

    venusian = venusian  # for testing injection

    def __init__(self, **settings):
        if 'for_' in settings:
            if settings.get('context') is None:
                settings['context'] = settings['for_']
        self.__dict__.update(settings)

    def __call__(self, wrapped):
        settings = self.__dict__.copy()

        def callback(context, name, ob):
            template = os.path.join(os.path.dirname(inspect.getfile(inspect.getmodule(ob))),
                                    settings.get('template'))
            if not os.path.isfile(template):
                raise ConfigurationError("No such file", template)

            contentType = settings.get('contentType', 'text/html')
            factory = WidgetTemplateFactory(template, contentType)
            provides = settings.get('provides', IPageTemplate)
            directlyProvides(factory, provides)

            config = context.config.with_package(info.module)
            config.registry.registerAdapter(factory,
                                            (settings.get('context', Interface),
                                             settings.get('layer', IRequest),
                                             settings.get('view', None),
                                             settings.get('field', None),
                                             settings.get('widget', ob)),
                                            provides,
                                            settings.get('mode', INPUT_MODE))

        info = self.venusian.attach(wrapped, callback, category='pyams_form')

        if info.scope == 'class':
            # if the decorator was attached to a method in a class, or
            # otherwise executed at class scope, we need to set an
            # 'attr' into the settings if one isn't already in there
            if settings.get('attr') is None:
                settings['attr'] = wrapped.__name__

        settings['_info'] = info.codeinfo  # fbo "action_method"
        return wrapped


def override_widget_template(widget, **settings):
    """Override template for a given widget

    This function can be used to override a widget template without using ZCML.
    Settings are the same as for @widgettemplate_config decorator.
    """
    stack = inspect.stack()[-1]
    template = os.path.join(os.path.dirname(inspect.getfile(inspect.getmodule(stack[0]))),
                            settings.get('template'))
    if not os.path.isfile(template):
        raise ConfigurationError("No such file", template)
    contentType = settings.get('contentType', 'text/html')
    factory = WidgetTemplateFactory(template, contentType)
    provides = settings.get('provides', IPageTemplate)
    directlyProvides(factory, provides)
    registry = getGlobalSiteManager()
    registry.registerAdapter(factory,
                             (settings.get('context', Interface),
                              settings.get('layer', IRequest),
                              settings.get('view', None),
                              settings.get('field', None),
                              widget),
                             provides,
                             settings.get('mode', INPUT_MODE))


class widgetlayout_config(object):
    """Class decorator used to declare a widget layout"""

    venusian = venusian  # for testing injection

    def __init__(self, **settings):
        if 'for_' in settings:
            if settings.get('context') is None:
                settings['context'] = settings['for_']
        self.__dict__.update(settings)

    def __call__(self, wrapped):
        settings = self.__dict__.copy()

        def callback(context, name, ob):
            template = os.path.join(os.path.dirname(inspect.getfile(inspect.getmodule(ob))),
                                    settings.get('template'))
            if not os.path.isfile(template):
                raise ConfigurationError("No such file", template)

            contentType = settings.get('contentType', 'text/html')
            factory = WidgetLayoutFactory(template, contentType)
            provides = settings.get('provides', IWidgetLayoutTemplate)
            directlyProvides(factory, provides)

            config = context.config.with_package(info.module)
            config.registry.registerAdapter(factory,
                                            (settings.get('context', Interface),
                                             settings.get('layer', IRequest),
                                             settings.get('view', None),
                                             settings.get('field', None),
                                             settings.get('widget', ob)),
                                            provides,
                                            settings.get('mode', INPUT_MODE))

        info = self.venusian.attach(wrapped, callback, category='pyams_form')

        if info.scope == 'class':
            # if the decorator was attached to a method in a class, or
            # otherwise executed at class scope, we need to set an
            # 'attr' into the settings if one isn't already in there
            if settings.get('attr') is None:
                settings['attr'] = wrapped.__name__

        settings['_info'] = info.codeinfo  # fbo "action_method"
        return wrapped


def override_widget_layout(widget, **settings):
    """Override layout template for a given widget

    This function can be used to override a widget layout template without using ZCML.
    Settings are the same as for @widgetlayout_config decorator.
    """
    stack = inspect.stack()[-1]
    template = os.path.join(os.path.dirname(inspect.getfile(inspect.getmodule(stack[0]))),
                            settings.get('template'))
    if not os.path.isfile(template):
        raise ConfigurationError("No such file", template)
    contentType = settings.get('contentType', 'text/html')
    factory = WidgetTemplateFactory(template, contentType)
    provides = settings.get('provides', IWidgetLayoutTemplate)
    directlyProvides(factory, provides)
    registry = getGlobalSiteManager()
    registry.registerAdapter(factory,
                             (settings.get('context', Interface),
                              settings.get('layer', IRequest),
                              settings.get('view', None),
                              settings.get('field', None),
                              widget),
                             provides,
                             settings.get('mode', INPUT_MODE))


#
# Reset button widget and action
#

@widgettemplate_config(mode=INPUT_MODE, template='templates/reset-input.pt', layer=IFormLayer)
@widgettemplate_config(mode=DISPLAY_MODE, template='templates/reset-display.pt', layer=IFormLayer)
@implementer_only(IResetWidget)
class ResetWidget(SubmitWidget):
    """A reset button of a form."""

    klass = u'reset-widget'
    css = u'reset'


@adapter_config(context=(IResetButton, IFormLayer), provides=IFieldWidget)
def ResetFieldWidget(field, request):
    reset = FieldWidget(field, ResetWidget(request))
    reset.value = field.title
    return reset


@adapter_config(context=(IFormLayer, IResetButton), provides=IButtonAction)
class ResetButtonAction(ResetWidget, ButtonAction):
    """Reset button action"""

    def __init__(self, request, field):
        Action.__init__(self, request, field.title)
        ResetWidget.__init__(self, request)
        self.field = field


#
# Close button widget and action
#

@widgettemplate_config(mode=INPUT_MODE, template='templates/close-input.pt', layer=IFormLayer)
@widgettemplate_config(mode=DISPLAY_MODE, template='templates/close-display.pt', layer=IFormLayer)
@implementer_only(ICloseWidget)
class CloseWidget(SubmitWidget):
    """A dialog close button"""

    klass = u'close-widget'
    css = u'close'


@adapter_config(context=(ICloseButton, IFormLayer), provides=IFieldWidget)
def CloseFieldWidget(field, request):
    close = FieldWidget(field, CloseWidget(request))
    close.value = field.title
    return close


@adapter_config(context=(IFormLayer, ICloseButton), provides=IButtonAction)
class CloseButtonAction(CloseWidget, ButtonAction):
    """Close button action"""

    def __init__(self, request, field):
        Action.__init__(self, request, field.title)
        CloseWidget.__init__(self, request)
        self.field = field


#
# Action button widget and action
#

@widgettemplate_config(mode=INPUT_MODE, template='templates/action-input.pt', layer=IFormLayer)
@widgettemplate_config(mode=DISPLAY_MODE, template='templates/action-display.pt', layer=IFormLayer)
@implementer_only(IActionWidget)
class ActionWidget(ButtonWidget):
    """A dialog action button"""

    label_css_class = FieldProperty(IActionWidget['label_css_class'])
    click_handler = FieldProperty(IActionWidget['click_handler'])
    url = FieldProperty(IActionWidget['url'])
    modal_target = FieldProperty(IActionWidget['modal_target'])


@adapter_config(context=(IActionButton, IFormLayer), provides=IFieldWidget)
def ActionFieldWidget(field, request):
    action = FieldWidget(field, ActionWidget(request))
    action.value = field.title
    return action


@adapter_config(context=(IFormLayer, IActionButton), provides=IButtonAction)
class ActionButtonAction(ActionWidget, ButtonAction):
    """Action button action"""

    def __init__(self, request, field):
        ActionWidget.__init__(self, request)
        Action.__init__(self, request, field.title)
        self.field = field
        self.label_css_class = field.label_css_class
        self.click_handler = field.click_handler
        self.url = field.url
        self.modal_target = field.modal_target


#
# Bytes field widget
#

@adapter_config(context=(IBytes, ITextAreaWidget), provides=IDataConverter)
class BytesFieldDataConverter(BaseDataConverter):
    """Bytes field data converter"""

    error_message = _("Invalid UTF-8 encoded data")

    def toWidgetValue(self, value):
        return value.decode('utf-8') if isinstance(value, bytes) else value

    def toFieldValue(self, value):
        return value.encode('utf-8') if isinstance(value, str) else value


#
# Integer field widget
#

SIGN_BEFORE_VALUE = (1, 3)
SIGN_AFTER_VALUE = (2, 4)


@adapter_config(context=(IInt, IIntegerWidget), provides=IDataConverter)
class IntegerFieldDataConverter(BaseDataConverter):
    """Integer field data converter"""

    error_message = _("Invalid integer value")

    def toWidgetValue(self, value):
        if value is self.field.missing_value:
            return ''
        return locale.format_string('%d', value, grouping=True)

    def toFieldValue(self, value):
        if not value:
            return self.field.missing_value
        try:
            return locale.atoi(value)
        except ValueError:
            raise FormatterValidationError(self.errorMessage, value)


@widgettemplate_config(mode=INPUT_MODE, template='templates/integer-input.pt', layer=IFormLayer)
@implementer_only(IIntegerWidget)
class IntegerWidget(TextWidget):
    """Integer widget"""

    @property
    def validate_rules(self):
        conv = locale.localeconv()
        allow_negative = True
        if (self.field.min is not None) and (self.field.min >= 0):
            allow_negative = False
        rules = {'pattern': '{}[\d{}]+{}'.format('\{}?'.format(conv.get('negative_sign', '-'))
                                                     if (allow_negative and
                                                         (conv.get('n_sign_posn') in SIGN_BEFORE_VALUE)) else '',
                                                 conv.get('thousands_sep', ''),
                                                 '\{}?'.format(conv.get('negative_sign', '-'))
                                                     if (allow_negative and
                                                         (conv.get('n_sign_posn') in SIGN_AFTER_VALUE)) else '')}
        return json.dumps(rules)


@adapter_config(context=(IInt, IFormLayer), provides=IFieldWidget)
def IntegerFieldWidget(field, request):
    return FieldWidget(field, IntegerWidget(request))


#
# Float field widget
#

@adapter_config(context=(IFloat, IFloatWidget), provides=IDataConverter)
class FloatFieldDataConverter(BaseDataConverter):
    """Float field data converter"""

    error_message = _("Invalid floating value")

    def toWidgetValue(self, value):
        if value is self.field.missing_value:
            return ''
        return locale.format_string('%.{0}f'.format(self.widget.precision), value, grouping=True)

    def toFieldValue(self, value):
        if not value:
            return self.field.missing_value
        try:
            return locale.atof(value)
        except ValueError:
            raise FormatterValidationError(self.error_message, value)


@widgettemplate_config(mode=INPUT_MODE, template='templates/float-input.pt', layer=IFormLayer)
@implementer_only(IFloatWidget)
class FloatWidget(TextWidget):
    """Float widget"""

    precision = FieldProperty(IFloatWidget['precision'])

    @property
    def validate_rules(self):
        conv = locale.localeconv()
        allow_negative = True
        if (self.field.min is not None) and (self.field.min >= 0):
            allow_negative = False
        rules = {'pattern': '{}[\d{}]+{}?\d*{}'.format('\{}?'.format(conv.get('negative_sign', '-'))
                                                           if (allow_negative and
                                                               (conv.get('n_sign_posn') in SIGN_BEFORE_VALUE)) else '',
                                                       conv.get('thousands_sep', ''),
                                                       conv.get('decimal_point', '\.'),
                                                       '\{}?'.format(conv.get('negative_sign', '-'))
                                                           if (allow_negative and
                                                               (conv.get('n_sign_posn') in SIGN_AFTER_VALUE)) else '')}
        return json.dumps(rules)


@adapter_config(context=(IFloat, IFormLayer), provides=IFieldWidget)
def FloatFieldWidget(field, request):
    return FieldWidget(field, FloatWidget(request))


#
# Date widget
#

@widgettemplate_config(mode=INPUT_MODE, template='templates/date-input.pt', layer=IFormLayer)
@implementer_only(IDateWidget)
class DateWidget(TextWidget):
    """Date widget"""


@adapter_config(context=(IDate, IFormLayer), provides=IFieldWidget)
def DateFieldWidget(field, request):
    return FieldWidget(field, DateWidget(request))


#
# Datetime widget
#

@adapter_config(context=(IDatetime, IDatetimeWidget), provides=IDataConverter)
class DatetimeDataConverter(BaseDatetimeDataConverter):
    """Datetime field data converter"""

    def toWidgetValue(self, value):
        value = tztime(value)
        return super(DatetimeDataConverter, self).toWidgetValue(value)

    def toFieldValue(self, value):
        value = super(DatetimeDataConverter, self).toFieldValue(value)
        return localgmtime(value)


@widgettemplate_config(mode=INPUT_MODE, template='templates/datetime-input.pt', layer=IFormLayer)
@implementer_only(IDatetimeWidget)
class DatetimeWidget(TextWidget):
    """Datetime widget"""


@adapter_config(context=(IDatetime, IFormLayer), provides=IFieldWidget)
def DatetimeFieldWidget(field, request):
    return FieldWidget(field, DatetimeWidget(request))


#
# Time widget
#

@widgettemplate_config(mode=INPUT_MODE, template='templates/time-input.pt', layer=IFormLayer)
@implementer_only(ITimeWidget)
class TimeWidget(TextWidget):
    """Time widget"""


@adapter_config(context=(ITime, IFormLayer), provides=IFieldWidget)
def TimeFieldWidget(field, request):
    return FieldWidget(field, TimeWidget(request))


#
# Color widget
#

@adapter_config(context=(IColorField, IColorWidget), provides=IDataConverter)
class ColorDataConverter(BaseDataConverter):
    """Color field data converter"""

    def toWidgetValue(self, value):
        value = super(ColorDataConverter, self).toWidgetValue(value)
        if value:
            value = '#' + value
        return value

    def toFieldValue(self, value):
        if value and value.startswith('#'):
            value = value[1:]
        return super(ColorDataConverter, self).toFieldValue(value)


@widgettemplate_config(mode=INPUT_MODE, template='templates/color-input.pt', layer=IFormLayer)
@implementer_only(IColorWidget)
class ColorWidget(TextWidget):
    """Color widget"""


@adapter_config(context=(IColorField, IFormLayer), provides=IFieldWidget)
def ColorFieldWidget(field, request):
    return FieldWidget(field, ColorWidget(request))


#
# HTML widget
#

@widgettemplate_config(mode=INPUT_MODE, template='templates/html-input.pt', layer=IFormLayer)
@widgettemplate_config(mode=DISPLAY_MODE, template='templates/html-display.pt', layer=IFormLayer)
@implementer_only(IHTMLWidget)
class HTMLWidget(TextAreaWidget):
    """HTML editor widget"""

    @property
    def editor_data(self):
        registry = self.request.registry
        config = registry.queryMultiAdapter((self.form, self.request), ITinyMCEConfiguration, name=self.name)
        if (config is None) and hasattr(self, 'basename'):  # I18n widget
            config = registry.queryMultiAdapter((self.form, self.request), ITinyMCEConfiguration, name=self.basename)
        if config is None:
            config = registry.queryMultiAdapter((self.form, self.request), ITinyMCEConfiguration)
        if config is None:
            config = getattr(self, 'editor_config_data', None)
            if config is not None:
                return json.dumps(config)
        else:
            return json.dumps(config.configuration)


@adapter_config(context=(IHTMLField, IFormLayer), provides=IFieldWidget)
def HTMLFieldWidget(field, request):
    return FieldWidget(field, HTMLWidget(request))


#
# Select2 widget
#

@widgettemplate_config(mode=INPUT_MODE, template='templates/select-input.pt', layer=IFormLayer)
@implementer_only(ISelect2Widget)
class Select2Widget(SelectWidget):
    """Select2 widget"""

    noValueMessage = _("(no selected value)")

    def get_content(self, entry):
        translate = self.request.localizer.translate
        return translate(entry['content'])


@adapter_config(context=(IChoice, IFormLayer), provides=IFieldWidget)
def ChoiceFieldWidget(field, request):
    return FieldWidget(field, Select2Widget(request))


@widgettemplate_config(mode=INPUT_MODE, template='templates/hidden-select-input.pt', layer=IFormLayer)
@implementer(IObjectData)
class HiddenSelect2Widget(Select2Widget):
    """Hidden select2 widget

    Widget values are provided through IObjectData
    """

    def extract(self, default=NO_VALUE):
        if self.name not in self.request and self.name + '-empty-marker' in self.request:
            return []
        value = self.request.get(self.name, default)
        if value != default:
            if not isinstance(value, (set, tuple, list)):
                value = set(value.split('|'))
            # do some kind of validation, at least only use existing values
            for token in value:
                if token == self.noValueToken:
                    continue
                try:
                    self.terms.getTermByToken(token)
                except LookupError:
                    return default
        return value

    @property
    def values(self):
        return '|'.join(self.value or '')

    @property
    def values_map(self):
        result = {}
        terms = self.terms
        [result.update({value: terms.getTermByToken(value).title}) for value in self.value or ()]
        return json.dumps(result)


def HiddenSelect2FieldWidget(field, request):
    return FieldWidget(field, HiddenSelect2Widget(request))


@adapter_config(context=(IList, HiddenSelect2Widget), provides=IDataConverter)
@adapter_config(context=(ITuple, HiddenSelect2Widget), provides=IDataConverter)
@adapter_config(context=(ISet, HiddenSelect2Widget), provides=IDataConverter)
class HiddenSelect2DataConverter(BaseDataConverter):
    """Hidden select2 data converter"""

    def toWidgetValue(self, value):
        return value or ()

    def toFieldValue(self, value):
        return set(value or ())


#
# TextLineList widget
#

@adapter_config(context=(ITextLineListField, ITextLineListWidget), provides=IDataConverter)
class TextLineListFieldDataConverter(BaseDataConverter):
    """Text line list field data converter"""

    def toWidgetValue(self, value):
        return '|'.join(value or [])

    def toFieldValue(self, value):
        return value.split('|') if value else None


@widgettemplate_config(mode=INPUT_MODE, template='templates/textlinelist-input.pt', layer=IFormLayer)
@widgettemplate_config(mode=DISPLAY_MODE, template='templates/textlinelist-display.pt', layer=IFormLayer)
@implementer_only(ITextLineListWidget)
class TextLineListWidget(TextWidget):
    """Text line list widget"""

    @property
    def tags(self):
        return json.dumps((self.value or '').split('|'))


@adapter_config(context=(ITextLineListField, IFormLayer), provides=IFieldWidget)
def TextLineListFieldWidget(field, request):
    return FieldWidget(field, TextLineListWidget(request))


#
# TextLine widget with SEO length status
#

@widgettemplate_config(mode=INPUT_MODE, template='templates/seo-textline-input.pt', layer=IFormLayer)
@implementer_only(ISEOTextLineWidget)
class SEOTextLineWidget(TextWidget):
    """SEO textline widget"""

    @property
    def length(self):
        return len(self.value or '')

    @property
    def status(self):
        status = 'success'
        length = self.length
        if length < 20 or length > 80:
            status = 'danger'
        elif length < 40 or length > 66:
            status = 'warning'
        return status


def SEOTextLineFieldWidget(field, request):
    return FieldWidget(field, SEOTextLineWidget(request))
